require_relative './linked_list.rb'

class LinkedList
  attr_reader :length

  def initialize
    @head = LinkedListNode.new
    @tail = LinkedListNode.new
    @head.next = @tail
    @tail.prev = @head

    @length = 0
  end

  def first
    @head.next == @tail ? nil : @head.next
  end

  def last
    @tail.prev == @head ? nil : @tail.prev
  end

  def append(value)
    node = value.class == LinkedListNode ? value : LinkedListNode.new(value)

    old_last = @tail.prev
    old_last.next = node
    @tail.prev = node
    node.prev = old_last
    node.next = @tail

    @length += 1
    node
  end

  def prepend(value)
    node = value.class == LinkedListNode ? value : LinkedListNode.new(value)

    old_first = @head.next
    old_first.prev = node
    @head.next = node
    node.next = old_first
    node.prev = @head

    @length += 1
    node
  end

  def pop
    return nil if @head.next == @tail
    remove(@tail.prev)
  end

  def shift
    return nil if @head.next == @tail
    remove(@head.next)
  end

  def remove(node)
    raise "Error in LinkedList#remove - given node is not a LinkedListNode, class is #{node.class}" unless node.class == LinkedListNode
    return if node == @tail || node == @head

    node.prev.next = node.next
    node.next.prev = node.prev
    node.prev = nil
    node.next = nil
    @length -= 1

    node
  end

  def each
    current = @head.next

    until current == @tail do
      yield current
      current = current.next
    end
  end

  def each_with_index
    current = @head.next
    counter = 0

    until current == @tail do
      yield current, counter
      current = current.next
      counter += 1
    end
  end

  def map(&prc)
    prc ||= Proc.new { |node| node }
    current = @head.next
    result = []

    until current == @tail do
      result << prc.call(current)
      current = current.next
    end

    result
  end

  def map_with_index(&prc)
    prc ||= Proc.new { |node, idx| node }
    current = @head.next
    counter = 0
    result = []

    until current == @tail do
      result << prc.call(current, counter)
      current = current.next
      counter += 1
    end

    result
  end
end
